# _*_ coding: utf-8 _*_
#
# hspylib v1.12.31
#
# Package: main.hspylib.core.config
"""Package initialization."""

__all__ = [
    'app_config', 
    'parser_factory', 
    'properties'
]
__version__ = '1.12.31'
